{-# LANGUAGE ScopedTypeVariables #-}

module Huffman where

import Data.Array (assocs, accumArray)
import Data.Bifunctor (first, second)
import Data.Char (isAscii, chr)
import Data.List (group, sort)

import LeftistHeap

newtype Bit = Bit Bool
  deriving (Eq)

instance Show Bit where
    show (Bit b) = case b of
        True  -> "1"
        False -> "0"

data HuffmanTree a = Tip a | Branch (HuffmanTree a) (HuffmanTree a)
    deriving Show

-- The number of occurrences of some symbol in some content to encode
type Frequency = Int

-- Associates each symbol with a bitstring.
type Codebook a = [(a, [Bit])]

frequencies :: Ord a => [a] -> [(a, Frequency)]
frequencies = undefined

frequenciesAscii :: String -> [(Char, Frequency)]
frequenciesAscii = undefined

huffmanTree :: Ord a => [(a, Frequency)] -> Maybe (HuffmanTree a)
huffmanTree = undefined

-- Create a codebook from a huffman tree.
codewords :: HuffmanTree a -> Codebook a
codewords (t :: HuffmanTree a) = case t of
    Tip x      -> [(x, [])]
    Branch l r -> codewordsL ++ codewordsR
      where
        codewordsL, codewordsR :: Codebook a
        codewordsL = map (second (Bit False :)) (codewords l)
        codewordsR = map (second (Bit True  :)) (codewords r)

-- Encode a symbol sequence with a codebook to a bit sequence.
-- Return Nothing if a symbol has no codebook entry.
encode :: Eq a => [a] -> Codebook a -> Maybe [Bit]
encode symbols codewords = concat <$> mapM (`lookup` codewords) symbols

-- Decode a bit sequence with a huffman tree.
-- Return Nothing if the tree is a single leaf or if there are undecodable 
-- leftover bits.
decode :: [Bit] -> HuffmanTree a -> Maybe [a]
decode _    (Tip _) = Nothing
decode bits tree    = sequence $ chop (decSymbol tree) bits
  where
    decSymbol :: HuffmanTree a -> [Bit] -> (Maybe a, [Bit])
    decSymbol (Branch l _) (Bit False:bits) = decSymbol l bits
    decSymbol (Branch _ r) (Bit True :bits) = decSymbol r bits
    decSymbol (Tip x)      bits             = (Just x, bits)
    decSymbol _            _                = (Nothing, [])

-- chop f l uses the splitting function f to repeatedly chop pieces from 
-- the list l. Returns a list of all pieces.
chop :: ([a] -> (b, [a])) -> [a] -> [b]
chop _ [] = []
chop f l  = b : chop f l'
  where
    (b, l') = f l